/**************************************************/
/* makefile.cpp - source for 'makefile'		  */
/**************************************************/

/**************************************************/
/* You should not need to touch anything here     */
/* unless you are adding a new platform. Change   */
/* makefile instead.				  */
/**************************************************/

/* "$Id: makefile.cpp,v 1.1.1.1 2002/10/23 16:55:38 bhepple Exp $" */

.SUFFIXES: .def .opn .dat
.dat.opn:
	./sm2c $<
.dat.def:
	./sm2c $<

LIBDIR=$(LOCAL)/lib
BINDIR=$(LOCAL)/bin
INCDIR=$(LOCAL)/include
MANDIR=$(LOCAL)/man/cat1

/* Some defaults - can be overridden below */
#ifdef MAXREC
MXR=-DSM_MAXREC=MAXREC
#endif

#ifdef MAKE_DEBUG_VERSION
DEBUG=-g -DSM_DEBUG=1
LDEBUG=-g
#else
DEBUG=-O -s
LDEBUG=
#endif

#ifdef DEBUG_MALLOC
DEBUG_HEAP=-DSM_DEBUG_HEAP=1
HEAP_OBJ=malloc$(OBJ)
#endif

#ifdef USE_GNU_CC
CC=gcc
#endif

CFLAGS=$(MXR) $(DEBUG) $(DEBUG_HEAP) $(CURSES) -I $(INCDIR) -D$(OSNAME)
LDFLAGS=$(LDEBUG) -L. -L $(LIBDIR)
EXTRA_SM2C_LIBS=$(HEAP_OBJ)

/* There's a space character here ... Unix likes it; TurboC hates it */
#define OUTPUT -o $@

OBJ=.o
EXE=
LIBSUFFIX=.a

RM=rm -f
INSTALL=cp

INCS=smio.h smlib.h

#ifdef USE_XAW
XOBJS=smxaw$(OBJ)
XINCS=smxaw.h
XOTHER=scrmgr.xct
SM_XDEF=-DSM_XAW=1
#endif

OTHERS=smdisp$(EXE) smedit$(EXE) smsort$(EXE) smgrep$(EXE) smfind$(EXE)
SCRIPTS=db2sm

/**************************************************/
/*** SunOS under System V curses ******************/
/**************************************************/

#ifdef SMSUNOS
OSNAME=SM_SUNOS
sun_default: all

ctool:
	cc $(DEBUG) ctool.c -o ctool -lsuntool -lsunwindow -lpixrect
xctool: xctool.c
	cc -O xctool.c -o xctool -I$(OPENWINHOME)/include -I$(INCDIR) -L$(OPENWINHOME)/lib -L$(LIBDIR) -lxview -lolgx -lX11
/* Note: xctool used to bomb if loaded -O instead of -g */

OTHERS=xctool$(EXE) smdisp$(EXE) smedit$(EXE) smsort$(EXE) smgrep$(EXE) smfind$(EXE)
XWINHOME=$(OPENWINHOME)

#ifdef USE_GNU_CC
CFLAGS=$(DEBUG) $(CURSES) $(MXR) -nostdinc -O -I . -I- -I /usr/5include -I /usr/include -I$(INCDIR) -DSM_SUNOS $(SM_XDEF)
LIBDIR=/usr/5lib
LDFLAGS=$(LDEBUG) -L$(LIBDIR)
LIBS=-L. -L$(LIBDIR) -lsm -L /usr/5lib/ -l$(CURSESLIB)  /usr/5lib/libc.a $(XLIBS)

#else /* Native compiler */

CC=/usr/5bin/cc

#ifdef USE_XVIEW
#ifdef USE_XAW
cant use both old chap!
#endif
XOBJS=smxview$(OBJ)
SM_XDEF=-DSM_XVIEW=1
XLIBS=-L$(OPENWINHOME)/lib -lxview -lolgx -lX11
smxview$(OBJ): smxview.c
	$(CC) -c $(CFLAGS) $(DEBUG) $(CURSES) -I$(OPENWINHOME)/include $(SM_XDEF) smxview.c
#endif /* XVIEW */

#ifdef SHARED_LIBS
LIBSUFFIX=.so.1.00
CFLAGS=$(DEBUG) $(CURSES) -pic -I$(INCDIR) -DSM_SUNOS $(SM_XDEF)
LIBS=-lsm -l$(CURSESLIB) $(XLIBS)
#else
CFLAGS=$(DEBUG) $(CURSES) -I$(INCDIR) -DSM_SUNOS $(SM_XDEF)
LIBS=libsm.a -L$(LIBDIR) -l$(CURSESLIB) $(XLIBS)
#endif /* SHARED_LIBS */

#endif /* !USE_GNU_CC */

#undef sun /* otherwise config.sun -> config.1! */

#endif /* SM_SUNOS */

/**************************************************/
/*** SCO Xenix 386 2.3.3 **************************/
/**************************************************/

#ifdef SMXENIX

OSNAME=SM_XENIX

#ifdef USE_GNU_CC
#if 0

This doesnt actually work ...!
#ifdef MAKE_DEBUG_VERSION
OTHEROBJ=$(LIBDIR)/GDB_space$(OBJ)
#endif

#endif

#endif /* !USE_GNU_CC */

LIBS=-L$(LIBDIR) -lsm -l$(CURSESLIB) -lmalloc -lx

#endif /* SM_XENIX */

/**************************************************/
/*** Generic System V - e.g. Linux ****************/
/**************************************************/

#ifdef SMSYSV

OSNAME=SM_SYSV

LIBS=-lsm -l$(CURSESLIB)
#define USE_NCURSES 1

#ifdef USE_XVIEW
#ifdef USE_XAW
cant use both old chap!
#endif
XOBJS=smxview$(OBJ)
SM_XDEF=-DSM_XVIEW=1
XLIBS=-L$(OPENWINHOME)/lib -lxview -lolgx -lX11
smxview$(OBJ): smxview.c
	$(CC) -c $(CFLAGS) $(DEBUG) $(CURSES) -I$(OPENWINHOME)/include $(SM_XDEF) smxview.c
#endif /* XVIEW */

#endif /* SM_SYSV */

/**************************************************/
/*** MS-DOS compiled by Turbo-C 2.0 ***************/
/**************************************************/

#ifdef SMTURBOC

OSNAME=SM_TURBOC

#undef OUTPUT
OBJ=.obj
EXE=.exe
LIBSUFFIX=.lib
OTHERS=smedit$(EXE) smsort$(EXE) smgrep$(EXE) smfind$(EXE)
SCRIPTS=
CC=tcc
INSTALL=copy
/* RM=del ... brain damaged rm - use MKS version or my own! */
/* No space here ... */
#define OUTPUT -o$@
#ifdef MAKE_DEBUG_VERSION
DEBUG=-v -y -N  /* No optimise; add debugging */
#else
DEBUG=-G -O -N- /* Optimise for speed */
#endif

/* -A to enable ANSI-C __STDC__ etc ... but turboc is too picky! */
/* -g0: stop after x warnings (i.e. don't stop for warnings */
CFLAGS=-ml -g0 $(DEBUG) $(CURSES) -I$(INCDIR) -DSM_TURBOC

LDFLAGS=-ml $(DEBUG) $(CURSES) -L$(LIBDIR)
LIBS=libsm.lib lcurses.lib
EXTRA_SM2C_LIBS=

#endif /* SM_TURBOC */

/**************************************************/
/*** Silicon Graphics under 4Sight ****************/
/**************************************************/

#ifdef SMSGI

OSNAME=SM_SGI

#ifndef USE_GNU_CC
CC=cc
#endif /* !USE_GNU_CC */

LIBS=-lsm -l$(CURSESLIB) 

#endif /* SM_SGI */

/**************************************************/
/* End of configurable variables 		  */
/**************************************************/

#ifdef USE_NCURSES
CURSES=-DUSE_NCURSES=1
CURSESLIB=ncurses
#else
CURSESLIB=curses
#endif

default: all

#ifdef USE_XAW
XLIBS=-L$(XWINHOME)/lib -lXaw -lXmu -lXt -lXext -lX11
smxaw$(OBJ): smxaw.c
	$(CC) -c $(CFLAGS) $(DEBUG) -I$(XWINHOME)/include $(SM_XDEF) smxaw.c
#endif

SMLIBS=smlib$(OBJ) smread$(OBJ) smdb$(OBJ) smio$(OBJ) smwrite$(OBJ) smfield$(OBJ) smscreen$(OBJ) smutil$(OBJ) smgetfl$(OBJ) smwhere$(OBJ) smbuilt$(OBJ) smmaps$(OBJ) $(XOBJS) $(HEAP_OBJ)

SMLIB=libsm$(LIBSUFFIX)
#ifdef SMTURBOC
$(SMLIB): $(SMLIBS)
	-$(RM) $@
	/* this may need to be done manually because of MS-DOGs 
	   limit of 120 characters per command line!!
	for %%i in ($?) do echo "+%%i & > smlib.tlb */
	tlib $@ @smlib.tlb
#else
#ifdef SHARED_LIBS
$(SMLIB): $(SMLIBS)
	/usr/bin/ld -o $@ -assert pure-text $(SMLIBS)
#else
$(SMLIB): $(SMLIBS)
	ar rcv $@ $?
	ranlib $@
#endif /* SHARED_LIBS */
#endif

/* Note that sm2c is needed in building the others and the library */
PROGS=smcreate$(EXE) smdata$(EXE) smquick$(EXE) sm2awk$(EXE) smupdate$(EXE) smproj$(EXE) smclean$(EXE) $(OTHERS)

ALLPROGS=sm2c$(EXE) $(PROGS) $(OTHERS) $(SCRIPTS)

OBJFILES= \
smlib$(OBJ) \
smread$(OBJ) \
smmaps$(OBJ) \
smwrite$(OBJ) \
smfield$(OBJ) \
smscreen$(OBJ) \
smdb$(OBJ) \
smutil$(OBJ) \
smgetfl$(OBJ) \
smwhere$(OBJ) \
smio$(OBJ) \
$(XOBJS) \
smstubio$(OBJ) \
sm2c$(OBJ) \
smquick$(OBJ) \
smupdate$(OBJ) \
sm2awk$(OBJ) \
smbuilt$(OBJ) \
smclean$(OBJ) \
smedit$(OBJ) \
smsort$(OBJ) \
smjoin$(OBJ) \
smgrep$(OBJ) \
smfind$(OBJ)

$(PROGS): $(SMLIB)

$(OBJFILES): smlib.h smio.h

#ifdef SHARED_LIBS
sm2c$(EXE): sm2c$(OBJ) smstubio$(OBJ) smlib$(OBJ) smread$(OBJ) smmaps$(OBJ)
	$(CC) $(LDFLAGS) OUTPUT sm2c$(OBJ) smstubio$(OBJ) smlib$(OBJ) smread$(OBJ) smmaps$(OBJ) $(EXTRA_SM2C_LIBS)
	$(MAKE) -f makefile.unx $(SMLIB)
	rm -f sm2c
	$(MAKE) -f makefile.unx y sm2c_shared

sm2c_shared:
	$(CC) $(LDFLAGS) -o sm2c sm2c$(OBJ) smstubio$(OBJ) $(EXTRA_SM2C_LIBS) -lsm -l$(CURSESLIB)
#else
sm2c$(EXE): sm2c$(OBJ) smstubio$(OBJ) smlib$(OBJ) smread$(OBJ) smmaps$(OBJ)
	$(CC) $(LDFLAGS) OUTPUT sm2c$(OBJ) smstubio$(OBJ) smlib$(OBJ) smread$(OBJ) smmaps$(OBJ) $(EXTRA_SM2C_LIBS)
#endif

smlib$(OBJ): smlib.h smio.h

smbuilt.def: smbuilt.dat sm2c$(EXE)
	./sm2c smbuilt.dat

smbuilt$(OBJ): smlib.h smio.h smbuilt.dat smbuilt.def

smquick$(EXE): smquick$(OBJ) smwrite$(OBJ) smdb$(OBJ) smstubio$(OBJ) 
	$(CC) $(LDFLAGS) OUTPUT smquick$(OBJ) smwrite$(OBJ) smdb$(OBJ) smstubio$(OBJ) $(LIBS)

smupdate$(EXE): smupdate$(OBJ) smstubio$(OBJ)
	$(CC) $(LDFLAGS) OUTPUT smupdate$(OBJ) smstubio$(OBJ) $(LIBS)

smcreate.def: smcreate.dat sm2c$(EXE)
	./sm2c smcreate.dat

smcreate$(OBJ): smlib.h smio.h smcreate.dat smcreate.def

smcreate$(EXE): smcreate$(OBJ) smio.h smlib.h
	$(CC) $(LDFLAGS) OUTPUT smcreate$(OBJ) $(LIBS)

smdata.def: smdata.dat sm2c$(EXE)
	./sm2c smdata.dat

smdata$(OBJ): smlib.h smio.h smdata.dat smdata.def

smdata$(EXE):  smdata$(OBJ)
	$(CC) $(LDFLAGS) OUTPUT smdata$(OBJ) $(LIBS)

smedit.def: smedit.dat sm2c$(EXE)
	./sm2c smedit.dat

smedit$(OBJ): smlib.h smio.h smedit.dat smedit.def

smedit$(EXE):  smedit$(OBJ)
	$(CC) $(LDFLAGS) OUTPUT smedit$(OBJ) $(LIBS)

smsort.def: smsort.dat sm2c$(EXE)
	./sm2c smsort.dat

smsort$(OBJ): smlib.h smio.h smsort.dat smsort.def

smsort$(EXE):  smsort$(OBJ)
	$(CC) $(LDFLAGS) OUTPUT smsort$(OBJ) $(LIBS)

smjoin.def: smjoin.dat sm2c$(EXE)
	./sm2c smjoin.dat

smjoin$(OBJ): smlib.h smio.h smjoin.dat smjoin.def

smjoin$(EXE):  smjoin$(OBJ)
	$(CC) $(LDFLAGS) OUTPUT smjoin$(OBJ) $(LIBS)

smgrep.def: smgrep.dat sm2c$(EXE)
	./sm2c smgrep.dat

smgrep$(OBJ): smlib.h smio.h smgrep.dat smgrep.def

smgrep$(EXE):  smgrep$(OBJ)
	$(CC) $(LDFLAGS) OUTPUT smgrep$(OBJ) $(LIBS)

smfind.def: smfind.dat sm2c$(EXE)
	./sm2c smfind.dat

smfind$(OBJ): smlib.h smio.h smfind.dat smfind.def

smfind$(EXE):  smfind$(OBJ)
	$(CC) $(LDFLAGS) OUTPUT smfind$(OBJ) $(LIBS)

sm2awk$(EXE): sm2awk$(OBJ) smstubio$(OBJ)
	$(CC) $(LDFLAGS) OUTPUT sm2awk$(OBJ) smstubio$(OBJ) $(LIBS) 

smproj.def: smproj.dat sm2c$(EXE)
	./sm2c smproj.dat

smproj$(OBJ): smlib.h smio.h smproj.def

smproj$(EXE): smproj$(OBJ)
	$(CC) $(LDFLAGS) OUTPUT smproj$(OBJ) $(LIBS)

smclean$(EXE): smclean$(OBJ) smstubio$(OBJ) smdb$(OBJ)
	$(CC) $(LDFLAGS) OUTPUT smclean$(OBJ) smdb$(OBJ) smstubio$(OBJ) $(LIBS)

#ifndef SMTURBOC
smdisp$(EXE): smdisp$(OBJ)
	$(CC) $(LDFLAGS) OUTPUT smdisp$(OBJ) -l$(CURSESLIB)
#endif

all: $(ALLPROGS)

dos_manuals install_man manuals:
	(cd man; $(MAKE) 'MANDIR=$(MANDIR)' 'INSTALL=$(INSTALL)' $@)

#ifdef SHARED_LIBS
install: all
	$(INSTALL) $(ALLPROGS) $(XOTHER) $(BINDIR)
	$(INSTALL) $(SMLIB) $(LIBDIR)
#else
install: all
	$(INSTALL) $(ALLPROGS) scrmgr.xct $(BINDIR)
#endif /* SHARED_LIBS */

lib:	$(SMLIB)

install_bin: $(SMLIB)
	$(INSTALL) $(SMLIB) $(LIBDIR)
	$(INSTALL) $(INCS) $(XINCS) $(INCDIR)

SRCFILES= \
ctool.c \
xctool.c \
scrmgr.xct \
sm2awk.c \
sm2c.c \
smclean.c \
smdata.c \
smcreate.c \
smedit.c \
smio.c \
smstubio.c \
smxview.c \
smxaw.c \
smlib.c \
smread.c \
smmaps.c \
smsort.c \
smjoin.c \
smgrep.c \
smfind.c \
smwrite.c \
smfield.c \
smscreen.c \
smutil.c \
smgetfl.c \
smwhere.c \
smproj.c \
smquick.c \
smupdate.c \
smdisp.c \
smdb.c \
smbuilt.c \
smio.h \
smlib.h \
smxaw.h \
smxview.h \
makefile \
makefile.cpp \
smlib.tlb \
malloc.c \
pcurse.c \
smcreate.dat \
smdata.dat \
smbuilt.dat \
smsort.dat \
smjoin.dat \
smgrep.dat \
smfind.dat \
smedit.dat \
smproj.dat \
db2sm

OTHERFILES=BUGS README COPYRIGHT

ALLFILES=$(SRCFILES) $(OTHERFILES)

#ifdef SMTURBOC
arc:
	arc u sm.arc $(ALLFILES) man
#else
tar:
	(cd man; $(MAKE) clean)
	tar cvf ../scrmgr.tar $(ALLFILES) man
#endif

clean:
	-$(RM) *$(OBJ)
	-$(RM) *~ 
	-$(RM) core*
	-$(RM) junk* 
	-$(RM) *.def 
	-$(RM) *.opn 
	-$(RM) $(SMLIB)
	-$(RM) sm2c$(EXE) $(PROGS) $(OTHERS) smmaps
	-$(RM) -r distrib
	-(cd man && $(MAKE) clean)

rcscheckin: $(ALLFILES)
	ci -l $?
	touch rcscheckin
	(cd man; $(MAKE) $@)

distrib: all
	if [ ! -d distrib ] ; then mkdir distrib; fi
	(cd distrib; for DIR in bin man man/cat1 dosman lib lib/scrmgr; \
	    do \
		if [ ! -d $$DIR ] ; then mkdir $$DIR; fi \
	    done)
	$(INSTALL) $(ALLPROGS) distrib/bin
	(cd distrib/bin; ln -f smedit smmore)
	$(INSTALL) terminfo.src distrib/lib/scrmgr
	(cd man; $(MAKE))
	(cd man; $(MAKE) manuals dosman)
	$(INSTALL) 'man/cat1/*' distrib/man/cat1
	$(INSTALL) 'man/dosman/*' distrib/dosman
	(cd distrib; tar cvf - bin man lib|compress -c >scrmgr.tar.Z)
	@echo "distrib/scrmgr.tar.Z is ready"
	

/*
# DO NOT DELETE THIS LINE -- make depend depends on it. */
