/*
    Contact info:
    bhepple@freeshell.org
    http://bhepple.freeshell.org/dcalc

    Copyright (C) 1991-2002 Bob Hepple

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.

*/
/* SCRMGR front end to sort(1) */

#include <stdio.h>
#include <getopt.h>
#include <setjmp.h>
#include <sys/types.h>		/* for stat */
#include <sys/stat.h>		/* for stat */
#include "smlib.h"
#include "smio.h"
#include "smsort.def"

CHAR *prog_rcsid = "$Id: smsort.c,v 1.3 2002/12/21 19:51:53 bhepple Exp $";

static struct {
    CHAR	*rev_string;
    CHAR 	last_filename[MAXPATHLEN + 1];
    SM_SCREEN	screen;
} global;

SM_FILE	screenset;

jmp_buf		NoMoreHeap;

static CHAR *usage_msg[] = {
"Provide an easy to use front end to sort(1).",
"",
"screenfile   : a SCRMGR screen definition file.",
"screenname   : a SCRMGR name in screenfile to use (default = first).",
NULL
};

#if !READ_DAT_AT_RUNTIME
static CHAR *smsort_dat[] = {
#include "smsort.dat"
NULL,
};
#endif

static void usage()
{
    CHAR **c;
    
    fprintf(stderr, "%s of %s.\n", global.rev_string, sm_program_name);
    fprintf(stderr, 
"Usage: %s [screenfile[:screenname]]\n", sm_program_name);
    
    c = usage_msg;
    while (*c) {
        fputs(*c, stderr);
        fputc('\n', stderr);
        c++;
    }
    exit(2);
}

void sm_error_exit ARG1(CHAR *, msg)
{
    sm_term();
    puts(msg);
    exit(1);
}

void re_fail ARG2(char *, msg, char, op)
{
    sm_message(msg);
}

static void update_n_paint ARG3(int, window, SM_FIELD, field, CHAR *, value)
{
    sm_fld_set(field, value);
    sm_paint_fld(window, field->c_value, field, 1, 0);
}

static COMMAND field_name_trigger ARG3(SM_FIELD, 	field, 
				       COMMAND, 	command, 
				       int, 		window)
{
    int		field_num, char_start, char_end;
    char	buf[40];

    if (global.screen) {
	if (*sm_fld_enq(field) &&
	    (sm_fld_open(global.screen, sm_fld_enq(field)) == NULL)) {
	    sm_message("No such field in this file");
	    if ((command != SM_DELETE_DATA_C) &&
		(command != SM_NO_SAVE_EXIT_C))
		return(0); /* suppress the move off this field */
	    else
		return(command);
	}
	if (sm_calc_sort_params(global.screen, sm_fld_enq(field), &field_num, 
				&char_start, &char_end) == 0) {
	    sprintf(buf, "%d", field_num);
	    update_n_paint(window, sm_fld_open(sort_scr, "field_num"), buf);
	    sprintf(buf, "%d", char_start);
	    update_n_paint(window, sm_fld_open(sort_scr, "char_start"), buf);
	    sprintf(buf, "%d", char_end);
	    update_n_paint(window, sm_fld_open(sort_scr, "char_end"), buf);
	    /* reset the cursor in the right place etc */
	    sm_paint_fld(window, field->c_value, field, 1, 0);
	}
    }
    return(command);
}

static COMMAND month_trigger ARG3(SM_FIELD, 	field, 
				  COMMAND, 	command, 
				  int, 		window)
{
    SM_SCREEN	scr = field->parent; /* == where_scr */
    SM_FIELD	fld;

    /* If this field goes to 'Y' then set numeric to 'N' */
    if (*sm_fld_enq(field) == 'Y') {
	fld = sm_fld_open(scr, "numeric");
	if (fld)
	    update_n_paint(window, fld, "N");
	/* reset the cursor in the right place etc */
	sm_paint_fld(window, field->c_value, field, 1, 0);
    }
    return(command);
}

static COMMAND numeric_trigger ARG3(SM_FIELD, 	field, 
				    COMMAND, 	command, 
				    int, 	window)
{
    SM_SCREEN	scr = field->parent;
    SM_FIELD	fld;

    /* If this field goes to 'Y' then set month to 'N' */
    if (*sm_fld_enq(field) == 'Y') {
	fld = sm_fld_open(scr, "month");
	if (fld)
	    update_n_paint(window, fld, "N");
	/* reset the cursor in the right place etc */
	sm_paint_fld(window, field->c_value, field, 1, 0);
    }
    return(command);
}

/* Exported to the SCRMGR package */
f_ptr sm_get_field_trigger ARG1(SM_FIELD, field)
{
    if (     strcmp(field->trigger_fn, "month_trigger") == 0)
	return(month_trigger);
    if (     strcmp(field->trigger_fn, "numeric_trigger") == 0)
	return(numeric_trigger);
    if (     strcmp(field->trigger_fn, "field_name_trigger") == 0)
	return(field_name_trigger);
    return(NULL);
}

static SM_SCREEN load_dat_file ARG2(SM_FILE, screenset, CHAR *, filename)
{
    CHAR	*screenname;
    SM_SCREEN	screen;
    int		file_was_read = 0;

    if (filename && *filename) {
	screenname = strchr(filename, ':');
	if (screenname)
	    *screenname++ = 0;

	if (sm_fopen(screenset, filename) == SM_ERROR)
	    sm_error("Error reading screen file");
	else {
	    /* Read sort keys across from the .dat file */
	    
	    if (screenset->separator == '~')
		sm_scrmgr_screens()->separator = '<';
	    
	    if (screenname)
		screen = sm_scr_open(screenset, screenname);
	    else if (sm_first_scr(screenset))
		screen = *sm_first_scr(screenset);
	    if (screen) {
		SM_SORTKEY	*skp = sm_first_sortkey(screen);
		CHAR		*buf = alloca(SM_MAXREC);
		int		field_num, char_start, char_end;
		
		if (skp && *skp)
		    while (skp && *skp) {
			sm_scr_init(sort_scr, 0);
			sm_fld_set(sm_fld_open(sort_scr, "input_file"), 
				   screen->filename);
			sm_fld_set(sm_fld_open(sort_scr, "separator"),
				   &screenset->separator);
			sm_fld_set(sm_fld_open(sort_scr, "field_name"), 
				   (*skp)->sortfield_name);
			
			if (sm_calc_sort_params(screen, 
						(*skp)->sortfield_name,
						&field_num,
						&char_start,
						&char_end) == 0) {
			    sprintf(buf, "%d", field_num);
			    sm_fld_set(sm_fld_open(sort_scr,"field_num"), buf);
			    sprintf(buf, "%d", char_start);
			    sm_fld_set(sm_fld_open(sort_scr,"char_start"),buf);
			    sprintf(buf, "%d", char_end);
			    sm_fld_set(sm_fld_open(sort_scr, "char_end"), buf);
			}
			
			sm_fld_set(sm_fld_open(sort_scr, "reverse"), 
				   (*skp)->reverse? "Y": "N");
			sm_fld_set(sm_fld_open(sort_scr, "ignore_case"), 
				   (*skp)->ignore_case? "Y": "N");
			sm_fld_set(sm_fld_open(sort_scr, "month"), 
				   (*skp)->month? "Y": "N");
			sm_fld_set(sm_fld_open(sort_scr, "numeric"), 
				   (*skp)->numeric? "Y": "N");
			sm_screen_to_data(buf, sort_scr);
			smdb_write(sort_scr, buf);
			skp = sm_next_sortkey(skp);
		    }
		else {
		    sm_scr_init(sort_scr, 0);
		    sm_fld_set(sm_fld_open(sort_scr, "input_file"), 
			       screen->filename);
		    sm_fld_set(sm_fld_open(sort_scr, "separator"),
			       &screenset->separator);
		    sm_screen_to_data(buf, sort_scr);
		    smdb_write(sort_scr, buf);
		}
		smdb_seek(sort_scr, (INDEX) 0);
	    }
	    file_was_read = 1;
	}
    }

    if (file_was_read) {
	sm_fld_prot(sm_fld_open(sort_scr, "field_num"), 1);
	sm_fld_prot(sm_fld_open(sort_scr, "char_start"), 1);
	sm_fld_prot(sm_fld_open(sort_scr, "char_end"), 1);
	sm_fld_prot(sm_fld_open(sort_scr, "field_name"), 0);
    } else {
	sm_fld_prot(sm_fld_open(sort_scr, "field_num"), 0);
	sm_fld_prot(sm_fld_open(sort_scr, "char_start"), 0);
	sm_fld_prot(sm_fld_open(sort_scr, "char_end"), 0);
	sm_fld_prot(sm_fld_open(sort_scr, "field_name"), 1);
	smdb_read_file(NULL, sort_scr, screenset->untabify);
	screen = NULL;
    }
    return(screen);
}

CALLBACK(load_file_done)
{
    if (command == SM_YES_C) {
	smdb_delete_data(sort_scr);
	sm_clear_out(0);
	sm_message("Loading ...");
	strcpy(global.last_filename, sm_get_prompt_value());
	sm_clear_out(0);
    	global.screen = load_dat_file(sm_user_screens(), global.last_filename);
	sm_message("");
    }
    return(command);
}

int lambda ARG2(int, command, SM_SCREEN, screen)
{
    switch (command) {
    default:
	break;
    }
    return(command);
}

static int build_sort_line ARG2(CHAR *, buf, SM_SCREEN, screen)
{
    CHAR	*key, c;
    int		field_num, char_start, char_end;
    struct stat	statbuf;

    strcpy(buf, "sort ");
    buf = strchr(buf, 0);

    key = smdb_seek(screen, (INDEX) 0);
    if (!key || !*key)
	return(1);

    sm_data_to_screen(screen, key, 0);
    c = *sm_fld_get(screen, "separator");
    if (c == 0) c = ' ';

    if (c > ' ') { /* Don't put in a -t option for ' ' */
	if (c == '\\')
	    sprintf(buf, "-t'\\\\' ", c);
	else if (c == 't')
	    sprintf(buf, "-t'\t' ");
	else
	    sprintf(buf, "-t'%c' ", c);
	buf = strchr(buf, 0);
    } else if (c != ' ') {
	sm_message("Bad separator character");
	return(1);
    }

    c = *sm_fld_get(screen, "unique");
    if (c == 'Y') {
	strcat(buf, "-u ");
	buf = strchr(buf, 0);
    }
    while (key) {
	key = smdb_get_current(screen);
	sm_data_to_screen(screen, key, 0);
	field_num = atoi(sm_fld_get(screen, "field_num"));
	char_start = atoi(sm_fld_get(screen, "char_start"));
	char_end = atoi(sm_fld_get(screen, "char_end"));
	if (field_num >= 0) {
	    if (char_start > 0)
		sprintf(buf, "+%d.%d ", 
			field_num, char_start);
	    else
		sprintf(buf, "+%d ", field_num);
	    buf = strchr(buf, 0);
	    if (char_end > 0)
		sprintf(buf, "-%d.%d", field_num, char_end);
	    else
		sprintf(buf, "-%d", field_num + 1);
	    if      (*sm_fld_get(screen, "ignore_case") == 'Y')
		strcat(buf, "f");
	    if      (*sm_fld_get(screen, "month") == 'Y')
		strcat(buf, "M");
	    else if (*sm_fld_get(screen, "numeric")     == 'Y')
		strcat(buf, "n");
	    if      (*sm_fld_get(screen, "reverse")     == 'Y')
		strcat(buf, "r");
	    strcat(buf, " ");
	    buf = strchr(buf, 0);
	}
	key = smdb_read_next(screen);
    }
    key = smdb_seek(sort_scr, (INDEX) 0);
    sm_data_to_screen(screen, key, 0);

    if (*sm_fld_get(screen, "input_file") == 0) {
	sm_message("You must fill in the input filename!");
	return(1);
    }

    if (stat(sm_fld_get(screen, "input_file"), &statbuf)) {
	sm_message("No such file!");
	return(1);
    }

    strcat(buf, sm_fld_get(screen, "input_file"));
    strcat(buf, " ");
    buf = strchr(buf, 0);
    return(0);
}

static void sort_file ARG1(SM_SCREEN, screen)
{
    CHAR	*command = alloca(SM_MAXREC);
    CHAR	*buf = command, *output;
    CHAR	tmp[100];
    int		is_pipe = 0;

    if ((output = sm_fld_get(screen, "output_file")) != NULL) {
	while (*output && isspace(*output)) 
	    output++;
	if (*output == '|') {
	    is_pipe = 1;
	    output++;
	}
    } else
	output = sm_fld_get(screen, "input_file");

#ifdef SM_TURBOC
    if (build_sort_line(buf, screen))
	return;

    buf = strchr(buf, 0);
    if (is_pipe)
	sprintf(buf, "| %s", output);
    else {
	sm_mktemp(tmp, "TXXXXXX");
	sprintf(buf, ">%s", tmp);
    }
#else    
    if (is_pipe)
	strcpy(buf, "(");
    else
	strcpy(buf, "(if ");

    buf = strchr(buf, 0);
    if (build_sort_line(buf, screen))
	return;

    buf = strchr(buf, 0);
    if (is_pipe)
	sprintf(buf, "| %s)", output);
    else {
	sm_mktemp(tmp, "TXXXXXX");
	sprintf(buf, ">%s; then cp %s %s; fi)", tmp, tmp, output);
    }
#endif

    if (smio_system(command, 1) == 0) {
#ifdef SM_TURBOC
	if (!is_pipe)
	    rename(tmp, output);
#endif    	
	sm_message("File is now sorted");
    }
    unlink(tmp);
}

static int save_command ARG2(CHAR *, filename, SM_SCREEN, screen)
{
    CHAR	*buf = alloca(SM_MAXREC);
    FILE	*file;
    int		retval = 0;

    if (backup_file(filename) < 0) {
	sprintf(buf, 
		"Can't make a backup file for \"%s\"", 
		filename);
	sm_message(buf);
    } 
    
    if ((file = fopen(filename, "w")) == NULL) {
	sm_file_error("Can't write to file: ", filename, errno);
	if (file)
	    fclose(file);
	retval = 1;
    } else {
	build_sort_line(buf, screen);
	fputs(buf, file);
	fputc('\n', file);
	fclose(file);
	sprintf(buf, "Command line saved in \"%s\".", filename);
	sm_message(buf);
    }
    return(retval);
}

static void display_command ARG1(SM_SCREEN, screen)
{
    CHAR	*buf = alloca(SM_MAXREC);

    build_sort_line(buf, screen);
    sm_message(buf);
}

static void view_output ARG1(SM_SCREEN, screen)
{
    CHAR	*buf = alloca(SM_MAXREC), *p;
    
    p = getenv("PAGER");
    if (p && *p)
	strcpy(buf, p);
    else
	strcpy(buf, "more");
    strcat(buf, " ");
    if (*sm_fld_get(screen, "output_file"))
	strcat(buf, sm_fld_get(screen, "output_file"));
    else if (*sm_fld_get(screen, "input_file"))
	strcat(buf, sm_fld_get(screen, "input_file"));
    else {
	sm_message("Nothing to look at yet!");
	return;
    }
    if (strchr(buf, '|'))
	sm_message("Can't look at a pipeline!");
    else
	smio_system(buf, 0);
}

CALLBACK(do_save_command)
{
    FILE	*file;
    CHAR	*filename;

    /* Always entered after sm_prompt */
    if (command == SM_YES_C) {
	filename = sm_get_prompt_value();

	if ((file = fopen(filename, "r")) != NULL) {
	    CHAR	*buf = alloca(100);

	    fclose(file);
	    sprintf(buf, "\"%s\" already exists. Overwrite it?", filename);
	    if (sm_confirm(buf) != SM_YES_C)
		return(SM_QUIT_C);
	}
	save_command(filename, sort_scr);
	command = SM_POPCALLBACK;
    }

    return(command);
}

void setup_save_command()
{
    CHAR	pwd[MAXPATHLEN + 1];

    GETCWD(pwd, MAXPATHLEN);
    sm_add_callback(do_save_command, NULL);
    sm_setup_prompt("Enter file for the command line (current directory is:)", 
		    pwd, 
		    NULL,
		    NULL);
}

CALLBACK(smsort_main_loop)
{
    switch(command) {
    case SM_LOAD_FILE_C:
	{
	    CHAR	pwd[MAXPATHLEN + 1];

	    sm_add_callback(load_file_done, NULL);
	    GETCWD(pwd, MAXPATHLEN);
	    sm_setup_prompt("Enter sort key file (current directory is:)", 
			    pwd,
			    global.last_filename,
			    NULL);
	    return(0);
	}	
	break;
	
    case SORT_FILE:
	sort_file(sort_scr);
	break;

    case DISPLAY_COMMAND:
	display_command(sort_scr);
	break;

    case VIEW_OUTPUT:
	view_output(sort_scr);
	break;

    case SAVE_COMMAND:
	setup_save_command();
	return(0);

    case SM_CLEAR_FILE_C:
	if (!smdb_file_changed(sort_scr) || 
	    (sm_confirm("Clear sort keys?") == SM_YES_C)) {
	    smdb_delete_data(sort_scr);
	    global.last_filename[0] = 0;
	    sm_message("File cleared");
	}
	break;
	
    case SM_EXIT_C:
    case SM_NO_SAVE_EXIT_C:
	return(command);
	break;

    default: /* Probably a return from one of the sub routines */
	break;
    }

    sm_setup_edit(SM_BASE_WINDOW, 
		  sort_scr, 
		  SM_EXIT_C, 
		  lambda, 
		  sort_skey,
		  global.screen);
    return(0);
}

int main ARG2(int, argc, CHAR **, argv)
{
    int			c;

    /* Temporary back door */
    if (setjmp(NoMoreHeap)) {
	alloca(0);
	sm_error_exit(sm_mem_error);
    }

    global.rev_string = "$Revision: 1.3 $";
    global.last_filename[0] = 0;

    sm_pre_init(argc, argv);

    /* Get rid of X options and enter X if appropriate */
    sm_init(1, argc, argv);

    while ((c = getopt(argc, argv, "?")) != EOF)
        switch (c) {
        default:  
	    /* Unfortunately, if we've entered X in sm_init, then any error 
	       messages disappear */
	    usage(); 
	    break;
        }
    
    if ((optind < argc) && (strlen(argv[optind]) != 0))
	/* Use first argument as sort file */
	strcpy(global.last_filename, argv[optind++]);

#ifdef SM_DEBUG
    smio_flushinp();
#endif

    screenset = sm_scrmgr_screens();

#if READ_DAT_AT_RUNTIME
    {
	CHAR	buf[MAXPATHLEN + 1];

	buf[0] = 0;
#if NEEDS_PATH_FOR_PROGS
	strcpy(buf, sm_program_directory);
#endif
	strcat(buf, "smsort.dat");
	if (sm_fopen(screenset, buf) == SM_ERROR)
	    sm_error_exit("Error reading smsort.dat!");
    }
#else
    if (sm_dbopen(screenset, smsort_dat) == SM_ERROR)
	sm_error_exit("Error reading internal data block!");
#endif

#include "smsort.opn"

    /* Back door for when we run out of heap - horrible structure, but
       what can one do ?*/

    if (setjmp(NoMoreHeap)) {
	alloca(0);

	/* Prevent infinite loop */
	if (setjmp(NoMoreHeap)) {
	    alloca(0);
	    sm_error_exit(sm_mem_error);
	}

	sm_message(sm_mem_error);

	if (setjmp(NoMoreHeap)) {
	    /* Oh boy - really serious! */
	    alloca(0);
	    sm_error_exit("Ran out of heap - file(s) were not saved! Sorry!");
	}

	alloca(0);

	if (setjmp(NoMoreHeap) == 0)
	    sm_term();
	exit(1);
    }

    global.screen = load_dat_file(sm_user_screens(), global.last_filename);

    /*************/
    /* Main loop */
    /*************/

    sm_message(global.rev_string);

    sm_add_callback(smsort_main_loop, NULL);	
    sm_setup_edit(SM_BASE_WINDOW, 
		  sort_scr, 
		  SM_EXIT_C, 
		  lambda, 
		  sort_skey,
		  global.screen);
    sm_main_loop();
    sm_term();
    exit(0);
}
